"""
Filename: constants.py
Author: Kuan-Hao Chao
Date: 2025-03-20
Description: Constants for OpenSpliceAI.
"""

CL_max=10000
# Maximum nucleotide context length (CL_max/2 on either side of the 
# position of interest)
# CL_max should be an even number

SL=5000
# Sequence length of SpliceAIs (SL+CL will be the input length and
# SL will be the output length)
#############################
# Global variable definition
############################## 
EPOCH_NUM = 10


