__author__ = "Kuan-Hao Chao"
__contact__ = "kuanhao.chao@gmail.com"
__copyright__ = "Copyright 2025, Kuan-Hao Chao, Alan Mao"
__credits__ = ["Kuan-Hao Chao", "Alan Mao"]
__date__ = "2025/03/21"
__deprecated__ = False
__email__ =  "kuanhao.chao@gmail.com"
__license__ = "GPL-3.0"
__maintainer__ = "developer"
__status__ = "Kuan-Hao Chao"
__version__ = "0.0.5"